
function avt::Init, ipaddress
    
    self._ipaddress = ipaddress
    
    catch, error
    if error eq 0 then begin
        avt_open, self->IpAddress()
    endif else begin
        return, 0
    endelse
    catch, /cancel
    
    return, 1
end

function avt::IpAddress
    return, self._ipaddress
end

pro avt::testfun
    catch, error
    if error eq 0 then begin
        ret = testfun()
    endif else begin
        if !error_state.name eq 'AVT_UNPLUGGED' then begin
            print, 'Camera unplugged cosa faccio?'    
        endif
        print, 'Error index: ', error
        print, 'Error message: ', !ERROR_STATE.MSG
    endelse
    catch, /cancel
end

;;;;;;;;;;;;;; Binning ;;;;;;;;;;;;;;

pro avt::SetBinning, binning
    self->SetAttribute, 'BinningX', binning
    self->SetAttribute, 'BinningY', binning
    sensorW = self->GetAttribute('SensorWidth')
    sensorH = self->GetAttribute('SensorHeight')
    self->setROI, [0,0,sensorW/binning, sensorH/binning]
end

function avt::Binning
    binX = self->GetAttribute('BinningX')
    binY = self->GetAttribute('BinningY')
    if binX ne binY then message, 'BinX should be equal to BinY'
    return, binX
end

;;;;;;;;;;;;;; ROI ;;;;;;;;;;;;;;

;
; roi is in the form [left, top, width, height]
; e.g.: on a GC1350 set [0,0, 1360, 1024] to read the whole frame 
;
pro avt::setROI, roi
    self->SetAttribute, 'RegionX', ulong(roi[0])
    self->SetAttribute, 'RegionY', ulong(roi[1])
    self->SetAttribute, 'Width', ulong(roi[2])
    self->SetAttribute, 'Height', ulong(roi[3])
end

function avt::ROI
    left    = self->GetAttribute('RegionX')
    top     = self->GetAttribute('RegionY')
    width   = self->GetAttribute('Width')
    height  = self->GetAttribute('Height')
    return, [left,top,width, height]
end


;;;;;;;;;;;;;;; Frame rate ;;;;;;;;;;;;;;;;;;

pro avt::SetFrameRate, frameRate
    self->SetAttribute, 'FrameRate', float(frameRate)
end

function avt::FrameRate
    framerate = self->GetAttribute('FrameRate')
    return, framerate
end

;;;;;;;;;;;;;;; Exposure time [s] ;;;;;;;;;;;;;;;;;;

pro avt::SetExposure, exposure
    self->SetAttribute, 'ExposureValue', ulong(exposure*1d6)
    ;avt_set_exposure, self->IpAddress(), long(exposure*1d6)
end

function avt::Exposure
    exposure_value = self->GetAttribute('ExposureValue')
    return, exposure_value / 1d6
end

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

pro avt::SetAttribute, attribute_name, value 
    avt_set_attribute, self->IpAddress(), attribute_name, value
end

function avt::GetAttribute, attribute_name
    return, avt_get_attribute(self->IpAddress(), attribute_name)
end

function avt::DumpAttributes, quiet=quiet
    catch, error
    if error eq 0 then begin
        names  = avt_list_attribute(self->IpAddress())
        values = strarr(n_elements(names))
        for i=0, n_elements(names)-1 do begin
            values[i] = string(self->GetAttribute(names[i]))
            if not keyword_set(quiet) then print, format='(%"%30s:%40s")', names[i], values[i]  
        endfor     
        return, {label: names, value:values}
    endif else begin
        print, 'manage exception'
    endelse
    catch, /cancel
end

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;
; n_frames  : number of frames to acquire.
; info      : info frame structure: NOTE this contains STATUS that has to be 0 for each frame
; timestamp : timestamp in sec (from when) of each frame  TODO BUGGED? seems the timestamp of last-1 frame
; filename  : write fits file (with header) to filename
;
function avt::GetFrames, n_frames, info=info, timestamp=timestamp, filename=filename 
    catch, error
    if error eq 0 then begin
        frames = avt_acquire(self->IpAddress(), n_frames, info)
        if arg_present(timestamp) then begin
            timestamp = (ulong64(info(*).timestamphi)*2ULL^32+ulong64(info(*).timestamplo) ) / double(self->getAttribute('TimeStampFrequency'))
        endif
        if keyword_set(filename) then begin
            attr = self->DumpAttributes(/quiet)
            ; create standard header
            mkhdr, header, frames, /extend
            ; add attributes using hierarch extension because standard trim labels to 8 characters
            for i=0, n_elements(attr.label)-1 do begin
                ;fxaddpar, header, attr.label[i], attr.value[i]
                line = string(format='(%"HIERARCH AVT.%-40s = %-24s")',attr.label[i], attr.value[i])
                header = [header, line]
            endfor 
            lineend = where(strmid(header,0,4) eq 'END ', cnt)
            header  =[header[0:lineend-1],header[lineend+1:*], string(format='(%"%-80s")','END')]
            writefits, filename, frames, header 
        endif 
    endif else begin
        catch, /cancel
        if !error_state.name eq 'AVT_UNPLUGGED' then begin
            print, 'Camera unplugged cosa faccio?'    
        endif else begin
            message, /reissue_last
        endelse
    endelse
    catch, /cancel
    return, frames
end

pro avt::Cleanup
    avt_close, self->IpAddress()
end

pro avt__define
    struct = { avt, $
        _ipaddress        : "", $
        _exposure         : 0d, $ ; exposure time [s]
        _framerate        : 0d, $ ; desired framerate [Hz]
        _binning          : 0,  $ ; binning
        _dummy            : 0   $
    }
end


pro test_avt
    camera_list = avt_list()
    print, camera_list

    ; open a camera. Put the right IP address in here
    cam = obj_new('AVT', '192.168.18.10')
    
    ; acquire full frame a bin 1
    cam->setBinning, 1 & cam->setFrameRate, 4 & cam->setExposure, 10d-3
    ret = cam->getFrames(10, timestamp=timestamp, info=info)
    for i=0,9 do tvscl, ret[*,*,i]
    help, info[0], /str

    ; timestamp : BUG?
    print, timestamp

    ; acquire full frame a bin 4
    cam->setBinning, 4 & cam->setFrameRate, 20 & cam->setExposure, 0.5d-3
    ret = cam->getFrames(10, timestamp=timestamp, info=info)
    for i=0,9 do tvscl, ret[*,*,i]
    help, info[0], /str

    ; acquire ROI 100x100 with top-left corner in [30,50]  a bin 1
    cam->setBinning, 1 & cam->setFrameRate, 50 & cam->setExposure, 10d-3
    cam->setROI, [30,50,100,100]
    ret = cam->getFrames(10, timestamp=timestamp, info=info)
    for i=0,9 do tvscl, ret[*,*,i]
    help, info[0], /str

    obj_destroy, cam
end
