
#include <stdio.h>
#include <stdlib.h>
#include <malloc.h>

#if defined(_WINDOWS) || defined(_WIN32) 
#define WIN32_LEAN_AND_MEAN
#include <Windows.h>
#include <Winsock2.h>
#include "PvAPI.h" 
#include "PvRegIO.h"
#endif

#if defined(_LINUX) || defined(_QNX) || defined(_OSX)
#include <unistd.h>
#include <time.h>
#include <signal.h>
#include <arpa/inet.h>
#include "PvApi.h" 
#include "PvRegIo.h"
#endif

#include "idl_export.h"


/* prototype for IDL_Load */
int IDL_Load( void );
//void __stdcall FrameDoneCB(tPvFrame* pFrame);
void  FrameDoneCB(tPvFrame* pFrame);


/*
 * Define message codes and their corresponding printf(3) format
 * strings. Note that message codes start at zero and each one is
 * one less that the previous one. Codes must be monotonic and
 * contiguous.
 */
static IDL_MSG_DEF msg_arr[] =
{
  {  "AVT_GENERIC",   "%N %s" },
  {  "AVT_UNPLUGGED",   "%NCamera was unplugged." },
};

/*
 *  Global ....
 */    
#define AVTPOOLSIZE 50
tPvHandle           _handle_pool[AVTPOOLSIZE];
char                _ip_pool[AVTPOOLSIZE][16];

//unsigned int        _TimeStampFrequency = 0;
//double*             _timestamp;
//unsigned int        _n_frames;

/*
 * The load function fills in this message block handle with the
 * opaque handle to the message block used for this module. The other
 * routines can then use it to throw errors from this block.
 */
static IDL_MSG_BLOCK _msg_block;

void poolInit()
{
    int i;
    for (i=0; i<AVTPOOLSIZE; i++){
        memset(_ip_pool[i],0,16);
        _handle_pool[i] = NULL;
    }
}

tPvHandle* poolGetFromIP(char* ipaddress)
{
    int i;
    for (i=0; i<AVTPOOLSIZE; i++){
        if (strncmp(ipaddress, _ip_pool[i], 15) == 0) return _handle_pool[i];
    }
    return NULL;
}

int poolAdd(char* ipaddress, tPvHandle* handle)
{
    char buf[128]; 
    int i;
    if (poolGetFromIP(ipaddress)) return -2;
    for (i=0; i<AVTPOOLSIZE; i++){
        if (_handle_pool[i] == NULL) {
            strncpy(_ip_pool[i], ipaddress, 15);
            memcpy(&(_handle_pool[i]), handle, sizeof(tPvHandle));
            return 0;
        } else {
            sprintf(buf, "Pool %d -> %s",i, _ip_pool[i]);
            IDL_Message(IDL_M_GENERIC, IDL_MSG_INFO, buf);
        }
    } 
    return -1;
}

void poolRemove(char* ipaddress)
{
    int i;
    for (i=0; i<AVTPOOLSIZE; i++){
        if (strncmp(ipaddress, _ip_pool[i], 15) == 0) {
            memset(_ip_pool[i],0,16);
            _handle_pool[i] = NULL;
        }
    }
}

//
// register exit handler with IDL_ExitRegister()
//
void avt_exit_handler(void) 
{ 
    /* Cleanup Code Here */ 
    //IDL_Message(IDL_M_GENERIC, IDL_MSG_INFO, "Exit handler");
    PvUnInitialize();
} 

#if defined(_LINUX) || defined(_QNX) || defined(_OSX)
void Sleep(unsigned int time)
{
    struct timespec t,r;
    
    t.tv_sec    = time / 1000;
    t.tv_nsec   = (time % 1000) * 1000000;    
    
    while(nanosleep(&t,&r)==-1)
        t = r;
}
#endif

/*
 *  TODO this is not needed!!!
 * Function:  FrameDoneCB()
 * Purpose:   Callback from PvAPI related to the post handling of a frame
 * Comments:  none
 */
//void __stdcall FrameDoneCB(tPvFrame* pFrame)
void FrameDoneCB(tPvFrame* pFrame)
{
    //unsigned long frame_idx;
    //unsigned int  frameSize;
    //char buf[128]; 
   // __int64 bit32;

    //bit32 = 4294967296;

    //frame_idx = (unsigned long)pFrame->Context[0];
    //frameSize = (unsigned int)pFrame->ImageBufferSize;
    //if (pFrame->Status == ePvErrSuccess) {
    //    if (frame_idx < _n_frames) {

            //_timestamp[frame_idx] = ( (double)(pFrame->TimestampHi)*bit32+ (double)(pFrame->TimestampLo) )   / _TimeStampFrequency;
            //_timestamp[frame_idx] = (__int64)(pFrame->TimestampHi)*2^32 + (__int64)(pFrame->TimestampLo);
    //    }
   // }
   ;
}


/* AVT version */

static void AvtVersion(int argc, IDL_VPTR *argv)
{
    unsigned long pMajor, pMinor; 
    char  buf[128]; 
    PvVersion(&pMajor, &pMinor); 
    sprintf(buf, "Version is %d %d", pMajor, pMinor);
    IDL_Message(IDL_M_GENERIC, IDL_MSG_INFO, buf);
}

/*
 * 
 */
static void AvtUnInitialize(int argc, IDL_VPTR *argv)
{
    PvUnInitialize();
}


/* Implementation of the TESTFUN IDL function */
static IDL_VPTR testfun(int argc, IDL_VPTR *argv)
{
    char  buf[128]; 
    sprintf(buf, "Code of %s is %d", "AVT_GENERIC", IDL_MessageNameToCode(_msg_block, "AVT_GENERIC") );
    IDL_Message(IDL_M_GENERIC, IDL_MSG_INFO, buf);
    sprintf(buf, "Code of %s is %d", "AVT_UNPLUGGED", IDL_MessageNameToCode(_msg_block, "AVT_UNPLUGGED") );
    IDL_Message(IDL_M_GENERIC, IDL_MSG_INFO, buf);
    

    IDL_MessageFromBlock(_msg_block, IDL_MessageNameToCode(_msg_block, "AVT_GENERIC"), IDL_MSG_LONGJMP, "pippo");
    IDL_MessageFromBlock(_msg_block, IDL_MessageNameToCode(_msg_block, "AVT_UNPLUGGED"), IDL_MSG_LONGJMP);
    return IDL_StrToSTRING("OK");
}

static IDL_STRUCT_TAG_DEF info_tags[] = {
    { (char*)"UNIQUEID", 0, (void *) IDL_TYP_LONG, 0},
    { (char*)"CAMERANAME", 0, (void *) IDL_TYP_STRING, 0},
    { (char*)"MODELNAME", 0, (void *) IDL_TYP_STRING, 0},
    { (char*)"PARTNUMBER", 0, (void *) IDL_TYP_STRING, 0},
    { (char*)"SERIALNUMBER", 0, (void *) IDL_TYP_STRING, 0},
    { (char*)"FIRMWAREVERSION", 0, (void *) IDL_TYP_STRING, 0},
    { (char*)"PERMITTEDACCESS", 0, (void *) IDL_TYP_LONG, 0},
    { (char*)"INTERFACEID", 0, (void *) IDL_TYP_LONG, 0},
    { (char*)"INTERFACETYPE", 0, (void *) IDL_TYP_LONG, 0},
    { (char*)"IPCONFIGMODE", 0, (void *) IDL_TYP_LONG, 0},
    { (char*)"IPADDRESS", 0, (void *) IDL_TYP_STRING, 0},
    { (char*)"IPSUBNET", 0, (void *) IDL_TYP_STRING, 0},
    { (char*)"IPGATEWAY", 0, (void *) IDL_TYP_STRING, 0},
    { 0, 0, NULL, 0 }
};

typedef struct AVTInfo_struct {
    IDL_LONG   UniqueId;
    IDL_STRING CameraName;
    IDL_STRING ModelName;
    IDL_STRING PartNumber;
    IDL_STRING SerialNumber;
    IDL_STRING FirmwareVersion;
    IDL_LONG   PermittedAccess;
    IDL_LONG   InterfaceId;
    IDL_LONG   InterfaceType;
    IDL_LONG   IpConfigMode;
    IDL_STRING IpAddress;
    IDL_STRING IpSubnet;
    IDL_STRING IpGateway;
} IDL_AVTInfo;


/*  
 * AvtList return a vector of structures containing camera's info.
 * 
 * INPUT 
 *    None
 *
 * KEYWORD
 *    None
 *  */
static IDL_VPTR AvtList(int argc, IDL_VPTR *argv)
{
    tPvCameraInfoEx list[50]; 
    
    tPvIpSettings   IPSettings;
    struct in_addr  addr;

    char            buf[128]; 
    unsigned long   numCameras;
    unsigned int    i; 
    void*           p;
    IDL_AVTInfo*    s_info;
    tPvErr          err;

    IDL_VPTR        struct_vect;       

    //err = PvInitialize();
    //if (err != ePvErrSuccess) {
    //    IDL_Message(IDL_M_GENERIC, IDL_MSG_INFO, "Failed to initialize PvAPI");
    //    return IDL_GettmpInt(-1);
    //}
    
    i=0; 
    while (PvCameraCount()==0 && i<10) {
        Sleep(250);
        i++;
    }

    numCameras   = PvCameraListEx(list, 50, NULL,sizeof(tPvCameraInfoEx));
    if (numCameras < 50) numCameras  += PvCameraListUnreachableEx( &(list[numCameras]), 50-numCameras, NULL,sizeof(tPvCameraInfoEx));

    sprintf(buf, "%d camera detected", numCameras); 
    IDL_MessageFromBlock(_msg_block, IDL_MessageNameToCode(_msg_block, "AVT_GENERIC"), IDL_MSG_INFO, buf);
    
    if (numCameras == 0) {
    //    PvUnInitialize();
        return IDL_GettmpInt(-1);
    }
   
    // Create the structure definition
    p = IDL_MakeStruct(0, info_tags);
    s_info =  (IDL_AVTInfo*)IDL_MakeTempStruct(p, 1, &numCameras, &struct_vect, IDL_TRUE);
    
    // Fill structure fields
    for (i=0; i<numCameras; i++) {
        err = PvCameraIpSettingsGet(list[i].UniqueId, &IPSettings);
        if (err==ePvErrSuccess) {
            s_info[i].IpConfigMode = IPSettings.ConfigMode;

            addr.s_addr = IPSettings.CurrentIpAddress;
            IDL_StrStore(&s_info[i].IpAddress, inet_ntoa(addr));
            
            addr.s_addr = IPSettings.CurrentIpSubnet;
            IDL_StrStore(&s_info[i].IpSubnet,  inet_ntoa(addr));
            
            addr.s_addr = IPSettings.CurrentIpGateway;
            IDL_StrStore(&s_info[i].IpGateway, inet_ntoa(addr));
        } 

        sprintf(buf, "%s [ID %u] [Name %s] [IP %s]", list[i].SerialNumber, list[i].UniqueId, list[i].CameraName, 
                IDL_STRING_STR(&s_info[i].IpAddress)
                ); 
        IDL_MessageFromBlock(_msg_block, IDL_MessageNameToCode(_msg_block, "AVT_GENERIC"), IDL_MSG_INFO, buf);
        
        s_info[i].UniqueId = list[i].UniqueId;
        IDL_StrStore(&s_info[i].CameraName, list[i].CameraName);
        IDL_StrStore(&s_info[i].ModelName, list[i].ModelName);
        IDL_StrStore(&s_info[i].PartNumber, list[i].PartNumber);
        IDL_StrStore(&s_info[i].SerialNumber, list[i].SerialNumber);
        IDL_StrStore(&s_info[i].FirmwareVersion, list[i].FirmwareVersion);
        s_info[i].PermittedAccess = list[i].PermittedAccess;
        s_info[i].InterfaceId = list[i].InterfaceId;
        s_info[i].InterfaceType, list[i].InterfaceType;

    }
    
    //PvUnInitialize();

    return struct_vect;
}

/*
 * Close Camera
 */
static void AvtClose(int argc, IDL_VPTR *argv)
{
    tPvErr     err;
    char  buf[256]; 
    char*           ipaddress;
    tPvHandle*      handle; 

    ipaddress = IDL_VarGetString(argv[0]);
    
    handle = poolGetFromIP(ipaddress);
    if (!handle) return;

    err = PvCameraClose(handle);
    if (err != ePvErrSuccess) {
        sprintf(buf, "PvCameraClose error %d", err); 
        IDL_Message(IDL_M_GENERIC, IDL_MSG_INFO, buf);
    } else IDL_Message(IDL_M_GENERIC, IDL_MSG_INFO, "Camera close done");
    poolRemove(ipaddress);
}

/*
 * Open Camera
 */
static void AvtOpen(int argc, IDL_VPTR *argv)
{
    tPvCameraInfoEx info; 
    unsigned long   numCameras; 
    tPvErr          err;
    char            buf[256]; 
    tPvUint32       packetsize;
    char*           ipaddress;
    tPvHandle       handle; 
    
    ipaddress = IDL_VarGetString(argv[0]);

    //numCameras = PvCameraListEx(&info, 1, NULL, sizeof(tPvCameraInfoEx)); 
 
    // Open the first camera found, if its not already open.  (See 
    // function reference for PvCameraOpen for a more complex example.) 
    //if ((numCameras == 1) && (info.PermittedAccess & ePvAccessMaster)) 
    //    PvCameraOpen(info.UniqueId, ePvAccessMaster, &_handle); 
    if (poolGetFromIP(ipaddress)) {
        sprintf(buf, "AvtOpen: %s already open", ipaddress); 
        IDL_Message(IDL_M_GENERIC, IDL_MSG_INFO, buf);
        return;
    } else {
        IDL_Message(IDL_M_GENERIC, IDL_MSG_INFO, "Camera not in pool");
    }

    // open camera by ip address
    err = PvCameraOpenByAddr( inet_addr(ipaddress), ePvAccessMaster, &handle); 
    if (err != ePvErrSuccess) {
        sprintf(buf, "PvCameraOpenByAddr error %d", err); 
        IDL_Message(IDL_M_GENERIC, IDL_MSG_LONGJMP, buf);
        return;
    }

    // Adjust packet size (jumbo) 
    err = PvCaptureAdjustPacketSize(handle,8228); 
    if (err != ePvErrSuccess) {
        sprintf(buf, "PvCaptureAdjustPacketSize error %d", err); 
        IDL_Message(IDL_M_GENERIC, IDL_MSG_INFO, buf);
    }
    // report packet size
    err = PvAttrUint32Get(handle, "PacketSize", &packetsize); 
    if (err != ePvErrSuccess) {
        sprintf(buf, "PvAttrUint32Get error %d", err); 
        IDL_Message(IDL_M_GENERIC, IDL_MSG_INFO, buf);
    } else {
        sprintf(buf, "PacketSize is %d", packetsize); 
        IDL_Message(IDL_M_GENERIC, IDL_MSG_INFO, buf);
    }

    // Set PixelFormat to Mono16 (or Mono12Packed?) 
    err = PvAttrEnumSet(handle, "PixelFormat", "Mono16"); 
    if (err != ePvErrSuccess) {
        sprintf(buf, "PixelFormat->Mono16 error %d", err); 
        IDL_Message(IDL_M_GENERIC, IDL_MSG_INFO, buf);
    } else IDL_Message(IDL_M_GENERIC, IDL_MSG_INFO, "PixelFormat set to Mono16");
    
    //add handle to pool of open camera
    if ( (err=poolAdd(ipaddress, &handle)) != 0) {
        sprintf(buf, "AvtOpen poolAdd error %d", err); 
        IDL_Message(IDL_M_GENERIC, IDL_MSG_LONGJMP, buf);
    }
}

//
// INPUT
//      IPAddress (string)
//      Binning 
//
// Same binning is applied to x and y
//
//static void AvtSetBinning(int argc, IDL_VPTR *argv)
//{
//    unsigned int    binning;
//    tPvErr          err;
//    char            buf[256]; 
//    char*           ipaddress;
//    tPvHandle*      handle; 
//
//    ipaddress = IDL_VarGetString(argv[0]);
//    
//    handle = poolGetFromIP(ipaddress);
//    if (!handle) return;
//
//    binning = IDL_LongScalar(argv[1]);
//
//    //  determine timestamp frequency Hz
//    err = PvAttrUint32Set(handle,"BinningX", binning);
//    err = err || PvAttrUint32Set(handle,"BinningY", binning);
//    if (err != ePvErrSuccess) {
//        sprintf(buf, "Set Binning error %d", err); 
//        IDL_Message(IDL_M_GENERIC, IDL_MSG_LONGJMP, buf);
//    } else {
//        sprintf(buf, "Binning set to %d", binning); 
//        IDL_Message(IDL_M_GENERIC, IDL_MSG_INFO, buf);
//    }
//}


//
// INPUT
//      IPAddress (string)
//      FrameRate [Hz]
//
// Only FixedRate is supported (no free-run).
//
//static void AvtSetFrameRate(int argc, IDL_VPTR *argv)
//{
//    float           framerate;
//    tPvErr          err;
//    char            buf[256]; 
//    char*           ipaddress;
//    tPvHandle*      handle; 
//
//    ipaddress = IDL_VarGetString(argv[0]);
//    
//    handle = poolGetFromIP(ipaddress);
//    if (!handle) return;
//
//    framerate = (float)IDL_DoubleScalar(argv[1]);
//
//    //  determine timestamp frequency Hz
//    err = PvAttrFloat32Set(handle,"FrameRate", framerate);
//    if (err != ePvErrSuccess) {
//        sprintf(buf, "Set Framerate error %d", err); 
//        IDL_Message(IDL_M_GENERIC, IDL_MSG_INFO, buf);
//    } else {
//        sprintf(buf, "Framerate set to %g", framerate); 
//        IDL_Message(IDL_M_GENERIC, IDL_MSG_INFO, buf);
//    }
//}

//
// INPUT
//      IPAddress (string)
//      ExposureValue [microseconds]
//
//
//static void AvtSetExposureValue(int argc, IDL_VPTR *argv)
//{
//    unsigned int    exposure;
//    tPvErr          err;
//    char            buf[256]; 
//    char*           ipaddress;
//    tPvHandle*      handle; 
//
//    ipaddress = IDL_VarGetString(argv[0]);
//    
//    handle = poolGetFromIP(ipaddress);
//    if (!handle) return;
//
//    exposure = IDL_LongScalar(argv[1]);
//
//    err = PvAttrUint32Set(handle,"ExposureValue", exposure);
//    if (err != ePvErrSuccess) {
//        sprintf(buf, "Set ExposureValue error %d", err); 
//        IDL_Message(IDL_M_GENERIC, IDL_MSG_INFO, buf);
//    } else {
//        sprintf(buf, "ExposureValue set to %d", exposure); 
//        IDL_Message(IDL_M_GENERIC, IDL_MSG_INFO, buf);
//    }
//}


static IDL_VPTR AvtGetAttribute(int argc, IDL_VPTR *argv)
{
    tPvErr              err;
    char                buf[256]; 
    char*               ipaddress;
    char*               label;
    tPvHandle*          handle; 
    tPvAttributeInfo    info;

    ipaddress = IDL_VarGetString(argv[0]);
    
    handle = poolGetFromIP(ipaddress);
    if (!handle) return IDL_GettmpInt(-1); 

    label = IDL_VarGetString(argv[1]);

    if( (err=PvAttrInfo(handle, label, &info)) == ePvErrSuccess )
    {
        if(info.Datatype != ePvDatatypeCommand && strcmp(info.Category,"/Stats"))
        {
            switch(info.Datatype)
            {           
                case ePvDatatypeString:
                {
                    char lValue[256];
                    // we assume here that any string value will be less than 256 characters
                    // long, which we may not be the case
                    if(!PvAttrStringGet(handle,label,lValue,128,NULL))
                        return IDL_StrToSTRING(lValue);
                    else
                        IDL_Message(IDL_M_GENERIC, IDL_MSG_LONGJMP, "PvAttrStringGet error");
                    break;                
                }
                case ePvDatatypeEnum:
                {
                    char lValue[256];
                    // we assume here that any string value will be less than 256 characters
                    // long, wich we may not be the case
                    if(!PvAttrEnumGet(handle,label,lValue,128,NULL))
                        return IDL_StrToSTRING(lValue);
                    else
                        IDL_Message(IDL_M_GENERIC, IDL_MSG_LONGJMP, "PvAttrEnumGet error");
                    break;
                }
                case ePvDatatypeUint32:
                {
                    tPvUint32 lValue;
                    
                    if(!PvAttrUint32Get(handle,label,&lValue))
                        return IDL_GettmpLong(lValue); 
                    else
                        IDL_Message(IDL_M_GENERIC, IDL_MSG_LONGJMP, "PvAttrUint32Get error");
                    break;
                }
                case ePvDatatypeFloat32:
                {
                    tPvFloat32 lValue;
                    IDL_VPTR   result;
                    
                    if(!PvAttrFloat32Get(handle,label,&lValue)) {
                        result = IDL_Gettmp();
                        result->type = IDL_TYP_FLOAT;
                        result->value.f = lValue;
                        return result;
                    }
                    else
                        IDL_Message(IDL_M_GENERIC, IDL_MSG_LONGJMP, "PvAttrFloat32Get error");
                    break;
                }
                default:
                    IDL_Message(IDL_M_GENERIC, IDL_MSG_LONGJMP, "AVT_GET_ATTRIBUTE: Unsupported Datatype");
            }
        }
    } else {
        if (err == ePvErrNotFound) {
            sprintf(buf, "Attribute %s does not exist", label); 
            IDL_Message(IDL_M_GENERIC, IDL_MSG_LONGJMP, buf);
        }
        if (err != ePvErrSuccess) {
            sprintf(buf, "PvAttrInfo error %d", err); 
            IDL_Message(IDL_M_GENERIC, IDL_MSG_LONGJMP, buf);
        }
    }
    return IDL_GettmpInt(-1); 
}



static void AvtSetAttribute(int argc, IDL_VPTR *argv)
{
    tPvErr              err;
    char                buf[256]; 
    char*               ipaddress;
    char*               label;
    tPvHandle*          handle; 
    tPvAttributeInfo    info;

    ipaddress = IDL_VarGetString(argv[0]);
    
    handle = poolGetFromIP(ipaddress);
    if (!handle) return IDL_GettmpInt(-1); 

    label = IDL_VarGetString(argv[1]);

    if( (err=PvAttrInfo(handle, label, &info)) == ePvErrSuccess )
    {
        if(info.Datatype != ePvDatatypeCommand && strcmp(info.Category,"/Stats"))
        {
            switch(info.Datatype)
            {           
                case ePvDatatypeString:
                {
                    char* lValue;
                    lValue = IDL_VarGetString(argv[2]);
                    if(!PvAttrStringSet(handle,label,lValue))
                        return;
                    else
                        IDL_Message(IDL_M_GENERIC, IDL_MSG_LONGJMP, "PvAttrStringSet error");
                    break;                
                }
                case ePvDatatypeEnum:
                {
                    char* lValue;
                    lValue = IDL_VarGetString(argv[2]);
                    if(!PvAttrEnumSet(handle,label,lValue))
                        return;
                    else
                        IDL_Message(IDL_M_GENERIC, IDL_MSG_LONGJMP, "PvAttrEnumSet error");
                    break;
                }
                case ePvDatatypeUint32:
                {
                    tPvUint32 lValue;
                    lValue = IDL_ULongScalar(argv[2]);
                    
                    if(!PvAttrUint32Set(handle,label,lValue))
                        return;
                    else
                        IDL_Message(IDL_M_GENERIC, IDL_MSG_LONGJMP, "PvAttrUint32Set error");
                    break;
                }
                case ePvDatatypeFloat32:
                {
                    //tPvFloat32 lValue;
                    double lValue;
                    IDL_VPTR   result;
                    lValue = IDL_DoubleScalar(argv[2]);
                    //lValue = argv[2]->value.f;
                    
                    if( (err=PvAttrFloat32Set(handle,label, (tPvFloat32)lValue)) == ePvErrSuccess) 
                        return;
                    else
            		sprintf(buf, "PvAttrFloat32Set error %d", err); 
            		IDL_Message(IDL_M_GENERIC, IDL_MSG_LONGJMP, buf);
                    break;
                }
                default:
                    IDL_Message(IDL_M_GENERIC, IDL_MSG_LONGJMP, "AVT_SET_ATTRIBUTE: Unsupported Datatype");
            }
        }
    } else {
        if (err == ePvErrNotFound) {
            sprintf(buf, "Attribute %s does not exist", label); 
            IDL_Message(IDL_M_GENERIC, IDL_MSG_LONGJMP, buf);
        }
        if (err != ePvErrSuccess) {
            sprintf(buf, "PvAttrInfo error %d", err); 
            IDL_Message(IDL_M_GENERIC, IDL_MSG_LONGJMP, buf);
        }
    }
    return; 
}




// return vector of string listing all the attributes name
static IDL_VPTR AvtListAttribute(int argc, IDL_VPTR *argv)
{
    
    tPvErr              err;
    char                buf[256]; 
    char*               ipaddress;
    char*               label;
    tPvHandle*          handle; 
    tPvAttributeInfo    info;
    IDL_VPTR            ret_var;
    IDL_STRING*         data;                
    tPvAttrListPtr      lAttrs;
    tPvUint32           lCount;
    tPvUint32           i;

    ipaddress = IDL_VarGetString(argv[0]);
    
    handle = poolGetFromIP(ipaddress);
    if (!handle) return IDL_GettmpInt(-1); 
 
    if( (err=PvAttrList(handle,&lAttrs,&lCount)) !=ePvErrSuccess ){
        sprintf(buf, "PvAttrInfo error %d", err); 
        IDL_Message(IDL_M_GENERIC, IDL_MSG_LONGJMP, buf);
    }
    
    data  = (IDL_STRING*)IDL_MakeTempVector(IDL_TYP_STRING, lCount, IDL_ARR_INI_ZERO, &ret_var);
    
    for(i=0;i<lCount;i++) IDL_StrStore(&data[i], (char*)(lAttrs[i]));
    
    return ret_var;
}



static IDL_STRUCT_TAG_DEF frame_tags[] = {
    { (char*)"STATUS", 0, (void *) IDL_TYP_LONG, 0},
    { (char*)"WIDTH", 0, (void *) IDL_TYP_ULONG, 0},
    { (char*)"HEIGHT", 0, (void *) IDL_TYP_ULONG, 0},
    { (char*)"REGIONX", 0, (void *) IDL_TYP_ULONG, 0},
    { (char*)"REGIONY", 0, (void *) IDL_TYP_ULONG, 0},
    { (char*)"FORMAT", 0, (void *) IDL_TYP_LONG, 0},
    { (char*)"BITDEPTH", 0, (void *) IDL_TYP_ULONG, 0},
    { (char*)"BAYERPATTERN", 0, (void *) IDL_TYP_LONG, 0},
    { (char*)"FRAMECOUNT", 0, (void *) IDL_TYP_ULONG, 0},
    { (char*)"TIMESTAMPLO", 0, (void *) IDL_TYP_ULONG, 0},
    { (char*)"TIMESTAMPHI", 0, (void *) IDL_TYP_ULONG, 0},
    { 0, 0, NULL, 0 }
};

typedef struct AVTframe_struct {
    IDL_LONG   Status;              // Status of this frame
    IDL_ULONG   Width;               // Image width
    IDL_ULONG   Height;              // Image height
    IDL_ULONG   RegionX;             // Start of readout region (left)
    IDL_ULONG   RegionY;             // Start of readout region (top)
    IDL_LONG   Format;              // Image format
    IDL_ULONG   BitDepth;            // Number of significant bits
    IDL_LONG   BayerPattern;        // Bayer pattern, if bayer format
    IDL_ULONG   FrameCount;          // Rolling frame counter
    IDL_ULONG   TimestampLo;         // Time stamp, lower 32-bits
    IDL_ULONG   TimestampHi;         // Time stamp, upper 32-bits
} IDL_AVTFrame;
   
//
// 
//
// INPUT
//      IPAddress (string)
//      n frames 
//
// OUTPUT
//      timestamp (optional)
//
//
static IDL_VPTR AvtAcquire(int argc, IDL_VPTR *argv)
{
    tPvErr      err;
    tPvFrame*   frames;
    char        buf[256]; 
    int         i,j;
    char        ok;
    unsigned int frameSize; 
    UCHAR*      data;
    unsigned int        n_frames;
                    
    unsigned int    width;
    unsigned int    height;
    IDL_MEMINT  dim[3];
    IDL_VPTR    var_data;
    IDL_VPTR    var_timestamp;

    void*               p;
    IDL_AVTFrame*       s_frame;
    IDL_VPTR            frame_struct;       
    
    char*           ipaddress;
    tPvHandle*      handle;     

    i=0;
    j=0;
    ok = 1;
    width = 0;
    height = 0;

    ////////////////////////// Parsing IDL ////////////////////////////
 
    ipaddress = IDL_VarGetString(argv[0]);
    n_frames = IDL_LongScalar(argv[1]);
    dim[2] = n_frames;

    // see if we are connected
    handle = poolGetFromIP(ipaddress);
    if (!handle) return IDL_GettmpInt(-1);
    
    // Determine width and height to allocate the data area.
    err = PvAttrUint32Get(handle,"Width",&width);
    err = err || PvAttrUint32Get(handle,"Height",&height);
    if (err != ePvErrSuccess) {
        sprintf(buf, "Get width/height error %d", err); 
        IDL_Message(IDL_M_GENERIC, IDL_MSG_INFO, buf);
    } else {
        sprintf(buf, "Get width/height done = [%d:%d]", width, height); 
        IDL_Message(IDL_M_GENERIC, IDL_MSG_INFO, buf);
    }

    dim[0] = width;
    dim[1] = height;
    data      = IDL_MakeTempArray(IDL_TYP_INT, 3, dim, IDL_ARR_INI_ZERO, &var_data);
    if (data == NULL) IDL_MessageFromBlock(_msg_block, IDL_MessageNameToCode(_msg_block, "AVT_GENERIC"), IDL_MSG_LONGJMP,"Cannot allocate array");

    //_timestamp = (double*)IDL_MakeTempArray(IDL_TYP_DOUBLE, 1, &n_frames, IDL_ARR_INI_ZERO, &var_timestamp);

    //  determine timestamp frequency Hz
    //err = PvAttrUint32Get(handle,"TimeStampFrequency",&_TimeStampFrequency);
    //if (err != ePvErrSuccess) {
    //    sprintf(buf, "Get TimeStampFrequency error %d", err); 
    //    IDL_Message(IDL_M_GENERIC, IDL_MSG_INFO, buf);
    //} else {
    //    sprintf(buf, "TimeStampFrequency is %d Hz", _TimeStampFrequency); 
    //    IDL_Message(IDL_M_GENERIC, IDL_MSG_INFO, buf);
    //}

    //////////////////////////// Camera Setting //////////////////////


    // 0) determine image buffer size
    err = PvAttrUint32Get(handle,"TotalBytesPerFrame",&frameSize);
    if (err != ePvErrSuccess) {
        sprintf(buf, "Get totalbytesperframe error %d", err); 
        IDL_Message(IDL_M_GENERIC, IDL_MSG_INFO, buf);
    } else {
        sprintf(buf, "Get totalbytesperframe done. %d", frameSize); 
        IDL_Message(IDL_M_GENERIC, IDL_MSG_INFO, buf);
    }

    // 1) allocate image buffer
    // 2) Fill out a tPvFrame
    frames = (tPvFrame*)malloc(sizeof(tPvFrame)*n_frames);
    for(i=0;i<n_frames;i++)
    {
        frames[i].Context[0] = (void*)i;
        frames[i].ImageBuffer =  data + i*frameSize;  //malloc(frameSize);
        frames[i].ImageBufferSize = frameSize;
    }
    sprintf(buf, "Allocated %u frames of %lu bytes each\n",n_frames,frameSize);
    IDL_Message(IDL_M_GENERIC, IDL_MSG_INFO, buf);
    
    // Set AcquisitionMode to continuous 
    err = PvAttrEnumSet(handle, "AcquisitionMode", "Continuous"); 
    if (err != ePvErrSuccess) {
        sprintf(buf, "AcquisitionMode->continuous error %d", err); 
        IDL_Message(IDL_M_GENERIC, IDL_MSG_INFO, buf);
    }
    IDL_Message(IDL_M_GENERIC, IDL_MSG_INFO, "AcquisitionMode set to continuous done");

    // Set trigger in fixed rate
    err = PvAttrEnumSet(handle, "FrameStartTriggerMode", "FixedRate"); 
    if (err != ePvErrSuccess) {
        sprintf(buf, "FrameStartTriggerMode->fixedrate error %d", err); 
        IDL_Message(IDL_M_GENERIC, IDL_MSG_INFO, buf);
    } else IDL_Message(IDL_M_GENERIC, IDL_MSG_INFO, "TriggerMode set to fixedrate");
    
    
    // Set Integration Time to 10 ms 
    //err = PvAttrUint32Set(handle, "ExposureValue", 10000); 
    //if (err != ePvErrSuccess) {
    //    sprintf(buf, "AcquisitionMode->continuous error %d", err); 
    //    IDL_Message(IDL_M_GENERIC, IDL_MSG_INFO, buf);
    //} else IDL_Message(IDL_M_GENERIC, IDL_MSG_INFO, "ExposureValue set to 10ms");
    
    // Start Capture
    err = PvCaptureStart(handle); 
    if (err != ePvErrSuccess) {
        sprintf(buf, "PvCaptureStart error %d", err); 
        IDL_Message(IDL_M_GENERIC, IDL_MSG_INFO, buf);
    }
    IDL_Message(IDL_M_GENERIC, IDL_MSG_INFO, "PvCaptureStart done");

    // Start acquisition
    err = PvCommandRun(handle, "AcquisitionStart"); 
    if (err != ePvErrSuccess) {
        sprintf(buf, "AcquisitionStart error %d", err); 
        IDL_Message(IDL_M_GENERIC, IDL_MSG_INFO, buf);
        PvCaptureEnd(handle); 
    }
    IDL_Message(IDL_M_GENERIC, IDL_MSG_INFO, "AcquisitionStart done");
    
    // loop until we have acquired all the frames
    while(j<n_frames && ok)
    {
        // enqueue up to 100 frames
        for(i=0;i<100 && j<n_frames;i++,j++)  ok = !PvCaptureQueueFrame(handle,&(frames[j]),FrameDoneCB);
    
        //IDL_Message(IDL_M_GENERIC, IDL_MSG_INFO, "Queued 10 frames: waiting for them to be done");

        // wait for the last enqueued frames to be done before enqueuing the next ones
        if(PvCaptureWaitForFrameDone(handle,&(frames[j-1]),PVINFINITE)) ok = 0;           
        
        //IDL_Message(IDL_M_GENERIC, IDL_MSG_INFO, "10 frames done");
    } 
    
    // TODO manage case ok==0 

    // Always check that tPvFrame->Status equals ePvErrSuccess, when a frame returned to you to ensure the data is valid.
    //sprintf(buf, "Frame done. Status:%d Width %d Height %d FrameCount %d", frame.Status, frame.Width, frame.Height, frame.FrameCount); 
    //IDL_Message(IDL_M_GENERIC, IDL_MSG_INFO, buf);
    
    
    if (argc > 2) {
        /////////////////////// FILL IDL FRAME INFO STRUCTURE //////////////////////////////////
        //
        // Create the structure definition
        p = IDL_MakeStruct(0, frame_tags);
        s_frame = (IDL_AVTFrame*)IDL_MakeTempStruct(p, 1, &n_frames, &frame_struct, IDL_TRUE);
        
        // Fill structure fields
        for (i=0; i<n_frames; i++) {
            s_frame[i].Status = frames[i].Status;
            s_frame[i].Width = frames[i].Width;
            s_frame[i].Height = frames[i].Height;
            s_frame[i].RegionX = frames[i].RegionX;
            s_frame[i].RegionY = frames[i].RegionY;
            s_frame[i].Format = frames[i].Format;
            s_frame[i].BitDepth = frames[i].BitDepth;
            s_frame[i].BayerPattern = frames[i].BayerPattern;
            s_frame[i].FrameCount = frames[i].FrameCount;
            s_frame[i].TimestampLo = frames[i].TimestampLo;
            s_frame[i].TimestampHi = frames[i].TimestampHi;
        }
    
        IDL_VarCopy(frame_struct, argv[2]);
    }
    /////////////////////////////////////////////////////////////////////////////////////////


    // Acquisition stop
    err = PvCommandRun(handle, "AcquisitionStop"); 
    if (err != ePvErrSuccess) {
        sprintf(buf, "AcquisitionStop error %d", err); 
        IDL_Message(IDL_M_GENERIC, IDL_MSG_INFO, buf);
    }
    IDL_Message(IDL_M_GENERIC, IDL_MSG_INFO, "AcquisitionStop done");

    // Capture End
    err = PvCaptureEnd(handle); 
    if (err != ePvErrSuccess) {
        sprintf(buf, "PvCaptureEnd error %d", err); 
        IDL_Message(IDL_M_GENERIC, IDL_MSG_INFO, buf);
    }
    IDL_Message(IDL_M_GENERIC, IDL_MSG_INFO, "CaptureEnd done");
    
    // dequeue frames
    err = PvCaptureQueueClear(handle); 
    if (err != ePvErrSuccess) {
        sprintf(buf, "PvCaptureQueueClear error %d", err); 
        IDL_Message(IDL_M_GENERIC, IDL_MSG_INFO, buf);
    }
    IDL_Message(IDL_M_GENERIC, IDL_MSG_INFO, "CaptureQueueClear done");
 
    // deallocate frames
    free(frames);
    
    
    // Return array
    return  var_data;
}




int IDL_Load(void)
{
    tPvErr errcode;

    /*
    * These tables contain information on the functions and procedures
    * that make up the TESTMODULE DLM. The information contained in these
    * tables must be identical to that contained in testmodule.dlm.
    */
    static IDL_SYSFUN_DEF2 function_addr[] = {
        { testfun, "TESTFUN", 0, IDL_MAXPARAMS, 0, 0},
        { (IDL_FUN_RET) AvtAcquire, "AVT_ACQUIRE", 2, 3, 0, 0},
        { (IDL_FUN_RET) AvtList, "AVT_LIST", 0, 0, 0, 0},
        { (IDL_FUN_RET) AvtGetAttribute, "AVT_GET_ATTRIBUTE", 2, 2, 0, 0},
        { (IDL_FUN_RET) AvtListAttribute, "AVT_LIST_ATTRIBUTE", 1, 1, 0, 0},
    };

    static IDL_SYSFUN_DEF2 procedure_addr[] = {
        { (IDL_FUN_RET) AvtVersion, "AVT_VERSION", 0, 0, 0, 0},
        { (IDL_FUN_RET) AvtOpen, "AVT_OPEN", 1, 1, 0, 0},
        { (IDL_FUN_RET) AvtClose, "AVT_CLOSE", 1, 1, 0, 0},
        //{ (IDL_FUN_RET) AvtSetBinning, "AVT_SET_BINNING", 2, 2, 0, 0},
        //{ (IDL_FUN_RET) AvtSetFrameRate, "AVT_SET_FRAMERATE", 2, 2, 0, 0},
        //{ (IDL_FUN_RET) AvtSetExposureValue, "AVT_SET_EXPOSURE", 2, 2, 0, 0},
        { (IDL_FUN_RET) AvtSetAttribute, "AVT_SET_ATTRIBUTE", 3, 3, 0, 0},
        { (IDL_FUN_RET) AvtUnInitialize, "AVT_UNINITIALIZE", 0, 0, 0, 0},
    };


    /*
    * Create a message block to hold our messages. Save its handle where
    * the other routines can access it.
    */
    if (!(_msg_block = IDL_MessageDefineBlock("AVT",
	    				   IDL_CARRAY_ELTS(msg_arr), msg_arr))) return IDL_FALSE;

    /*
    * Register our routine. The routines must be specified exactly the same
    * as in testmodule.dlm.
    */
    if (!IDL_SysRtnAdd(function_addr, TRUE, IDL_CARRAY_ELTS(function_addr))) return IDL_FALSE;
    if (!IDL_SysRtnAdd(procedure_addr, FALSE, IDL_CARRAY_ELTS(procedure_addr))) return IDL_FALSE;

    /*
     *  Initialize PvAPI
     */  
    errcode = PvInitialize();
    if (errcode != ePvErrSuccess) {
        IDL_Message(IDL_M_GENERIC, IDL_MSG_INFO, "Failed to initialize PvAPI");
        return IDL_FALSE;
    }
    
    /*
     * Initialize connection pool
     */
    poolInit();

    /*
     *  Register exit handler
     */
    IDL_ExitRegister(avt_exit_handler);

    return IDL_TRUE;
}
