
;+
; INPUT 
;       SDK_DIR  Full path of folder containing the AVT GigE SDK. This should have inc-pc and lin-pc has subfolders
; 
; 1)
; PvApi.h has to be modified for use with cl VC98
; // LB 02 sep 2010
; //typedef long long       tPvInt64;   // 64-bit signed integer
; typedef __int64       tPvInt64;   // 64-bit signed integer
; //end LB 02 sep 2010
; 
; 2) Copy PvAPI.dll in windows/system32 (and maybe use regsrv32?)
;
;-
pro build_avt_dlm, VERBOSE=verbose, SDK_DIR

  ; Locate the avt_dlm files
  thisfile = routine_info("build_avt_dlm", /source)
  avt_dlm_dir = file_dirname(thisfile.path)

  case strlowcase(!version.os_family) of 
    'unix': begin
               extracflags =  '-D_LINUX -D_x86 -I"'+filepath(root=SDK_DIR,'inc-pc')+'"'
               extralflags =  '"'+filepath(root=SDK_DIR, sub=['lib-pc','x86','4.4'], 'libPvAPI.a')+'" -lstdc++'
               loader = 'gcc -shared -o %L %O %X'
	     end
    'windows':begin
               extracflags =  '/I"'+filepath(root=SDK_DIR,'inc-pc')+'"'
               extralflags =  '"'+filepath(root=SDK_DIR, sub=['lib-pc'], 'PvAPI.lib')+'" Ws2_32.lib'
             end
     else: message, 'Unknown OS type'
  endcase

  ; Build the avt DLM in the !MAKE_DLL.COMPILE_DIRECTORY directory.
  ;
  MAKE_DLL, ['avt_dlm'], 'avt_dlm','IDL_Load', INPUT_DIR=avt_dlm_dir, $
	VERBOSE=verbose, SHOW_ALL_OUTPUT=verbose, NOCLEANUP=verbose, $
        EXTRA_CFLAGS=extracflags, EXTRA_LFLAGS=extralflags, LD=loader, COMPILE_DIR=avt_dlm_dir
 
  ; NOTE TO IDL USERS: This call to DLM_REGISTER is not the recommended
  ; way to make DLMs known to your IDL session. The best, and usually
  ; the only recommended, way to do this is to put the DLM files into
  ; one of the directories searched by IDL at startup (and contained in
  ; the !DLM_PATH system variable). Not only is is easier, but IDL will
  ; know the routines from your DLM before it compiles a single line of
  ; PRO code, avoiding one of the most common LINKIMAGE pitfalls.
  ; DLMs are largely intended to provide a safer easier interface to
  ; the capabilities of LINKIMAGE, so this is an important point.
  ;
  ; We do this here in order to simplify the demo. Otherwise, it would be
  ; necessary to alter the !DLM_PATH, and restart IDL, none of which is
  ; really central to our demo.
  ;
  ;DLM_REGISTER, new_dlm_file
  print, 'Now you must add '+ avt_dlm_dir +' to the IDL_DLM_PATH and restart'

  ; Display the DLM so the user can see that it is there
  ;
  ;print,''
  ;HELP,/DLM, 'testmodule'

  ; Run the 2 routines. Note that I have to use EXECUTE so that IDL
  ; will see these as system routines within this BUILD_TESTMODULE
  ; routine. This is because BUILD_TESTMODULE was compiled by IDL
  ; before it knew about the TESTMODULE DLM, and as such, plain calls
  ; get turned into IDL user routine calls. Since EXECUTE compiles at
  ; runtime, it will see them as system routines now that DLM_REGISTER
  ; has been called.
  ;
  ; This is the LINKIMAGE pitfall mentioned above, and a very good
  ; reason not to make a habit of using DLM_REGISTER.
  ;
  ;void = execute('testpro')
  ;void = execute('print, testfun()')

end
